/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998-2000  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * VTun additions for SecoNET gateways.
 *
 * SecoNet 04/2015
 *   Generic wrapper functions for AES en/decryption.
 *   Used in auth.c (ECB mode) and lfd_encrypt.c (GCM mode)
 */ 

/*
   * Uses AES and RNG implementations from the polarSSL kit:
   * https://tls.mbed.org/
   * Copyright (C) 2006-2012, Brainspark B.V. All rights reserved.
   * Lead Maintainer: Paul Bakker <polarssl_maintainer at polarssl.org>
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <time.h>
//#include <ctype.h>

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include "vtun.h"
#include "linkfd.h"


#ifdef HAVE_SSL

#include "SecoNet_AES.h"

// Error code to be retrieved with AES_GetError()
static int f_lastError = 0;

// Crypto key and contexts:

	// AES key, will be set based on user config file from main.c at SecoNet:3
	// We set a (true random) default here, which gets used in case no valid key
	// or password is provided, to make sure we cannot end up using a zeroed out
	// AES key. We recommend quitting the application instead with an error
	// log message, and have the user fix the config file.
	// However, this decision must be made on the system level.
	// Note: This key was generated on www.random.org
static uint8 f_keyAES[ AESGCM_KEY_LENGTH ] =
{
	0xc3, 0x27, 0xd6, 0x3b, 0x15, 0x09, 0xdb, 0x7e, 0x4c, 0x45, 0xcd, 0x42, 0xb4, 0xae, 0x16, 0x3e
};
static TCtxAES f_ctxAES; // AES context, initialized with AES key in AES_Init()
static TCtxGCM f_ctxGCM; // As above, but for AES-GCM

//#define DEBUG_ENCRYPTION //Used during testing

	// If encryption debugging is enabled, encryptions are logged.
	// Size, IV, and key, are logged in one common file; payloads (original and
	// encryptes) are logged into a separate file for each encryption.
#ifndef DEBUG_ENCRYPTION
// Provide empty macros for the normal case where encryption debugging is disabled
#define _DEBUG_LogStart( )
#define _DEBUG_LogEnd( )
#define _DEBUG_LogHeader( sKey, sIV, length )
#define _DEBUG_LogData( ch, buffer, length )

#else // DEBUG_ENCRYPTION
// Note this define should never be enabled in production systems, as it would
// waste memory even if not actually used (no -d in the command line).
#warning "##################################################"
#warning "Encryption debugging is active! (DEBUG_ENCRYPTION)"
#warning "##################################################"
#define DEBUG_ENC_LOGCOUNT 10 // Number of encryptions that are logged
#define DEBUG_ENC_DIR "/root/" // Target directory for encryption log files
//#define DEBUG_ENC_DIR  VTUN_STAT_DIR // Target directory for encryption log files
static BOOL f_bDebugAES = FALSE; // flag to switch debugging on or off
//static uint8 f_aDebugKey[ AESGCM_KEY_LENGTH ];
//static uint8 f_aDebugIV[ AESGCM_KEY_LENGTH ];
static const char* f_sDebugLogName = DEBUG_ENC_DIR "vtund-aes.log"; // trace file: key, IV, and size
static char f_sDebugLogDataName[ 50 ] = DEBUG_ENC_DIR "vtund-aes.bin.TN"; // payload files
static FILE* f_pFileDebugLog = NULL; // trace file handle
static int f_iDebugLog = -1; // Index of encryption log output

void _Debug_LogInit( void );
static inline void _DEBUG_LogStart( void );
static inline void _DEBUG_LogEnd( void );
static inline void _DEBUG_LogHeader( uint8* sKey, uint8* sIV, int length );
static inline void _DEBUG_LogData( char ch, uint8* aBuffer, int length );

#endif // DEBUG_ENCRYPTION



/* Generic helper functions ***************************************************/

int AES_GetError( void ) { return f_lastError; }

#if !defined(AES_BLOCKSIZE_BITS) || AES_BLOCKSIZE_BITS != 128
#error "Key generation needs to be adapted for block size"
#endif

	//FIX: Perhaps use strtol here
static inline char HexcharToVal( char ch )
{
	//return ( ch & 0x0f ) + ( ch >> 6 ) * 9; // short but unsafe
	if( ch >= '0' && ch <= '9' ) return ch & 0x0f;
	ch |= 0x20u;
	if( ch >= 'a' && ch <= 'f' ) return ch - 'a' + 10;
	return -1;
}

	// Parse a string of 32 hex digits into the AES key
BOOL AES_SetKeyHex( const char* sKeyHex )
{
	char chHigh, chLow;
	int iByte;

	for( iByte = 0; iByte < AESGCM_KEY_LENGTH; iByte++ )
	{
		chHigh = HexcharToVal( sKeyHex[ iByte * 2 ] );
		if( chHigh < 0 )
			return FALSE;
		chLow = HexcharToVal( sKeyHex[ iByte * 2 + 1 ] );
		if( chLow < 0 )
			return FALSE;
		f_keyAES[ iByte ] = ( chHigh << 4 ) | chLow;
	}
	return iByte >= MINIMUM_AESKEY_LENGTH;
}

	// Generate a 128 bit key by md5-hashing the given password
BOOL AES_SetKeyPassword( const char* sPassword )
{
	md5( (uint8*)sPassword, strlen( sPassword ), f_keyAES );
	return TRUE;
}



/* Random number generation ***************************************************/

	// Initialize RNG context with an arbitrary personalization string
	// (and possibly add entropy sources).
	// Used in auth.c to generate a random challenge for authentification.
static inline int Rnd_Init( ctr_drbg_context* pRndContext )
{
	char ps[ 9 ] = { 0xab, 0x9f, 0x4b, 0x91, 0x27, 0xc8, 0xc0, 0x1c, '\0' }; // Arbitrary string for personalization
		// NOTE: possibly individualize by adding ID/IP/MAC or similar.

	entropy_context ec;
		// TODO: Check if entropy sources are available

	entropy_init( &ec );
	
	f_lastError = ctr_drbg_init( pRndContext, entropy_func, &ec, (unsigned char *) ps, strlen( ps ) );
	
	//entropy_free( &ec ); // Only used in later versions of polarSSL

	return f_lastError;
}


	// Generate a random byte sequence.
	// NOTE: Perhaps initialize only one drbg context for the module?
int RAND_bytes( uint8* bytes, int count )
{
   ctr_drbg_context rndCtx;

   f_lastError = Rnd_Init( &rndCtx );
   if( f_lastError != 0 ) return f_lastError;

   f_lastError = ctr_drbg_random( &rndCtx, bytes, count );

   //ctr_drbg_free( &rndCtx ); // Only used in later versions of polarSSL

   return f_lastError;
}


/* AES en/decryption **********************************************************/

// Some functions here are 1:1 wrappers provided only in case further
// modifications/additions are necessary. If not, these should be inlined.

	// Initialize the AES module for encryption.
int AES_InitEncrypt( void )
{
	f_lastError = aes_setkey_enc( &f_ctxAES, f_keyAES, AES_KEYSIZE_BITS);
	//memset( f_keyAES, 0, AES_KEY_LENGTH );

	return f_lastError;
}

	// Initialize for decryption.
int AES_InitDecrypt( void )
{
	f_lastError = aes_setkey_dec( &f_ctxAES, f_keyAES, AES_KEYSIZE_BITS );
	//memset( f_keyAES, 0, AES_KEY_LENGTH );
	return f_lastError;
}


	// Cleanup of AES module.
	// Erase key if used temporary
void AES_Exit( void )
{
	//memset( f_aeskey, 0, AESGCM_KEY_LENGTH );
}


	// Encrypt a plaintext block with AES-ECB. Bitsize has been set to 128 in f_ctxAES.
	// Used during initial authentication (auth.c)
int AES_EncryptECB( char* aInput, char* aOutput )
{
	f_lastError = aes_crypt_ecb( &f_ctxAES, AES_ENCRYPT, (uint8*)aInput, (uint8*)aOutput );
	return f_lastError;
}

	// Decrypt a plaintext block with AES-ECB. Bitsize is set in f_ctxAES to 128.
	// Used during initial authentication (auth.c)
int AES_DecryptECB( char* aInput, char* aOutput )
{
	f_lastError = aes_crypt_ecb( &f_ctxAES, AES_DECRYPT, (uint8*)aInput, (uint8*)aOutput );
	return f_lastError;
}


/* AES-GCM en/decryption ******************************************************/

	// As AES_Init(), but for GCM mode.
int AESGCM_Init( void )
{
	_DEBUG_LogStart();
	//HashKey( key, keyLength );
	f_lastError = gcm_init( &f_ctxGCM, f_keyAES, AESGCM_KEYSIZE_BITS );

/* For now, we set the key only once and don't clear it between uses.
#ifndef DEBUG_ENCRYPTION
	// Debug-version needs the key later in _Debug_LogOutput, so not cleared here.
	memset( f_keyAES, 0, AESGCM_KEY_LENGTH ); // Paranoia
#endif
*/
	return f_lastError;
}

	// Cleanup of GCM module.
void AESGCM_Exit( void )
{
	_DEBUG_LogEnd( );
/*
#ifdef DEBUG_ENCRYPTION
	memset( f_keyAES, 0, AESGCM_KEY_LENGTH );
#endif // DEBUG_ENCRYPTION
*/
}


// Helper functions to retrieve pointers to various elements in the header:

static inline size_t LengthAAD( void ) { return AESGCM_AAD_LENGTH; }
static inline size_t LengthTAG( void ) { return AESGCM_TAG_LENGTH; }
static inline size_t LengthIV( void ) { return AESGCM_IV_LENGTH; }
static inline size_t LengthTrail( void ) { return AESGCM_BUFFER_TRAIL; }

static inline size_t NetLength( int length ) { return length - AESGCM_HEADER_LENGTH; }
   // Net length of packet excluding the authentication header

   // The payload VTUN hands to us consists of our encryption header followed by
   // the tunneled packet. Header consists of AAD, TAG, and IV.
static inline uint8* BufferAAD( char* buffer ) {
   return (uint8*)buffer; }
static inline uint8* BufferTAG( char* buffer ) {
   return BufferAAD( buffer ) + LengthAAD( ); }
static inline uint8* BufferIV( char* buffer ) { // NOTE: Use 12 payload bytes as IV?
   return BufferTAG( buffer ) + LengthTAG( ); }
static inline uint8* BufferInput( char* buffer ) {
   return BufferIV( buffer ) + LengthIV( ); }

   // Output buffer is in different locations for encoding and decoding:
   // When encoding, the input buffer can be re-used (see gcm.h)
static inline uint8* BufferEncOutput( char* buffer ) {
   return BufferInput( buffer ); }

   // When decoding, the IV is copied out beforehand, and part of the space is
   // used for the output trailing offset of 8 bytes.
static inline uint8* BufferDecOutput( char* buffer ) {
   return BufferInput( buffer ) - LengthTrail( ); }

   // Pointer to the struct containing the additional authenticated data.
static inline TAddData* AAD_Pointer( char* buffer ) {
   return (TAddData*)BufferAAD( buffer ); }

   // Set the sequence number in the AAD struct.
static inline void AAD_SetSeqNumber( char* buffer, uint16 num ) {
   AAD_Pointer( buffer )->seqNum = num; }


   // Encrypt the data in the given buffer (in place) with the AES mode set
   // in f_ctxGCM (128 bit).
   // buffer holds AES header in addition to the usual VTUN packet.
   // length is the size of the total message including AES header, in bytes
   // output will be set to buffer, as we enrypt in-place.
int AESGCM_Encrypt( char* buffer, int length, char** pOutput )
{
   static uint16 s_seqNum = 0;
   // Create a random initialization vector and put it into the message buffer,
   // so it is transmitted to the receiver along with the data
   f_lastError = RAND_bytes( BufferIV( buffer ), LengthIV( ) );
   if( f_lastError != 0 ) return f_lastError;

   _DEBUG_LogHeader( f_keyAES, BufferIV( buffer ), length );
   _DEBUG_LogData( 'p', BufferInput( buffer ), length ); // Log plaintext data

   AAD_SetSeqNumber( buffer, s_seqNum++ ); // Set the current sequence number
   *pOutput = buffer; // The entire buffer including overhead needs to be transmitted

	f_lastError = gcm_crypt_and_tag(
		&f_ctxGCM, GCM_ENCRYPT, NetLength( length ), // length of input to be encrypted
		BufferIV( buffer ), LengthIV( ), // initialization vector and length (in bytes),
		BufferAAD( buffer ), LengthAAD( ), // additional data and length (bytes)
		BufferInput( buffer ),   // input pointer
		BufferEncOutput( buffer ),  // output pointer
		LengthTAG( ), BufferTAG( buffer ) ); // tag and length (bytes)

	_DEBUG_LogData( 'e', BufferEncOutput( buffer ), length );

	// On error, return zero length; otherwise, entire length is used, since the
	// AES header is sent along with the payload.
	return f_lastError != 0 ? 0 : length;
}

	// Decrypt the data in the given buffer (using the key set in f_ctxGCM). We
	// use only one buffer, so the output pointer must be offset by 8 bytes ahead
	// of the input (see comments in gcm.h); BufferOutput() takes care of that.
	// Returns the length of the decrypted buffer (excluding the header)
	// buffer holds AES header and payload.
	// length is the length of the total message including header, in bytes.
	// pDecoded will be set to point to the output buffer (the actual packet)
	// TODO: Might want to check the sequence number.
int AESGCM_Decrypt( char* buffer, int length, char** pDecoded )
{
	char initVect[ AESGCM_IV_LENGTH ];

	// Read initialization vector from message (space is re-used for buffer trailing)
	memcpy( initVect, BufferIV( buffer ), LengthIV( ) );

	*pDecoded = (char*)BufferDecOutput( buffer ); // after decoding, we only use the payload

	f_lastError = gcm_crypt_and_tag(
		&f_ctxGCM, GCM_DECRYPT, NetLength( length ), // context, mode=decrypt, length of input buffer
		BufferIV( buffer ), LengthIV( ), // initialization vector and length (in bits),
		BufferAAD( buffer ), LengthAAD( ), // additional data and length
		BufferInput( buffer ),   // input pointer
		BufferDecOutput( buffer ),  // output pointer
		LengthTAG( ), BufferTAG( buffer ) ); // tag and length

	// Again zero length on error, otherwise the length of the payload only, which
	// is the package that is sent out on the local interface.
	return f_lastError != 0 ? 0 : length - AESGCM_HEADER_LENGTH;
}


	// Same as above, but checks the tag to authenticate the message.
int AESGCM_DecryptAuth( char* buffer, int length, char** pDecoded )
{
	uint8 initVect[ AESGCM_IV_LENGTH ];

	// Read initialization vector from message (space is re-used for buffer trailing)
	memcpy( initVect, BufferIV( buffer ), LengthIV( ) );

	*pDecoded = (char*)BufferDecOutput( buffer ); // after decoding, we only use the payload

	f_lastError = gcm_auth_decrypt(
		&f_ctxGCM, NetLength( length ),
		initVect, LengthIV( ),
		BufferAAD( buffer ), LengthAAD( ), // additional data and length
		BufferTAG( buffer ), LengthTAG( ), // Tag received
		BufferInput( buffer ),   // input pointer
		(uint8*)*pDecoded );  // output pointer

	return f_lastError != 0 ? 0 : NetLength( length );
}


/* Encryption debug logging functions *****************************************/

/* If enabled, DEBUG_ENC_LOGCOUNT log files will be created:
 * vtund-aes.bin.pX (X=A,B,...) contains the plaintext data
 * vtund-aes.bin.eX (X=A,B,...) the corresponding encrypted data
 * vtund-aes.log will log the key and IVs used
 */


#ifdef DEBUG_ENCRYPTION

// Called from main when -debug switch is detected
void _Debug_LogInit( void )
{
	f_bDebugAES = TRUE;
	unlink( f_sDebugLogName );
	vtun_syslog( LOG_INFO,"Encryption debugging enabled." );
}

	// Open the log files for writing (unless count is already exceeded)
static inline void _DEBUG_LogStart( void )
{
	if( !f_bDebugAES )
		return;

	if( f_iDebugLog >= DEBUG_ENC_LOGCOUNT )
		f_pFileDebugLog = NULL;
	else
	{
		f_pFileDebugLog = fopen( f_sDebugLogName, "a" );
		vtun_syslog( LOG_INFO,"Opened log file %s, file=%p.", f_sDebugLogName, f_pFileDebugLog );
	}
}

	// Close the log files (if open)
static inline void _DEBUG_LogEnd( void )
{
	if( !f_bDebugAES )
		return;
	if( f_pFileDebugLog != NULL )
	{
		vtun_syslog( LOG_INFO,"Closing log file %p.", f_pFileDebugLog );
		fclose( f_pFileDebugLog );
		f_pFileDebugLog = NULL;
	}
}


static inline void BytesToHex( char* sBuffer, uint8* aBytes, size_t count )
{
	size_t i;

	for( i = 0; i < count; i++ )
	{
		sprintf( sBuffer + 2 * i, "%02hx", aBytes[ i ] );
	}
}

static inline void _DEBUG_LogHeader( uint8* sKey, uint8* sIV, int length )
{
	FILE* pFile;
	uint8 sFullIV[ AES_KEY_LENGTH ];
	char sHexKey[ AES_KEY_LENGTH * 2 + 1 ];
	char sHexIV[ AES_KEY_LENGTH * 2 + 1 ];
	char sBuffer[ 120 ];

	if( !f_bDebugAES|| f_iDebugLog >= DEBUG_ENC_LOGCOUNT )
		return;

	f_iDebugLog++;
	memset( sFullIV, 0, AES_KEY_LENGTH );
	memcpy( sFullIV, sIV, LengthIV( ) );
	sFullIV[ 15 ] = 1; // Set counter part of IV to 1.

	BytesToHex( sHexKey, sKey, AES_KEY_LENGTH );
	BytesToHex( sHexIV, sFullIV, AES_KEY_LENGTH );

	pFile = fopen( f_sDebugLogName, "a" );
	//pFile = f_pFileDebugLog; // use this if log file is opened/closed once with the above functions
	if( pFile != NULL )
	{
		sprintf( sBuffer, "Enc %c: size=%d, key='%s', IV='%s'\n",
					'A' + f_iDebugLog, length, sHexKey, sHexIV );
		fputs( sBuffer, pFile );
		fclose( pFile );
	}
}

	// Adjust filename, open binary file, and write the input payload
static inline void _DEBUG_LogData( char ch, uint8* aBuffer, int length )
{
	FILE* pFile;
	char* cp;

	if( !f_bDebugAES || f_iDebugLog >= DEBUG_ENC_LOGCOUNT )
		return;

	cp = f_sDebugLogDataName + strlen( f_sDebugLogDataName ) - 2 ;
	cp[ 0 ] = ch;
	cp[ 1 ] = 'A' + f_iDebugLog;

	pFile = fopen( f_sDebugLogDataName, "wb" );
	if( pFile != NULL)
	{
		fwrite( aBuffer, 1, NetLength( length ), pFile );
		fclose( pFile );
	}
}

#endif // DEBUG_ENCRYPTION



#endif /* HAVE_SSL */
